import argparse
from . import job
import subprocess
import os
import pathlib
import warnings

from . import utils as ut
from . import config as cfg


def custom_formatwarning(msg, *args, **kwargs):
    # ignore everything except the message
    return ut.cstring(str(msg) + '\n', 'black_yellowbg')


warnings.formatwarning = custom_formatwarning


def submit_func(uargs):
    '''
    Wrapper for CLI submit call

    Parameters
    ----------
    uargs : argparser object
        User arguments

    Returns
    -------
    None

    '''

    if len(ut.get_envvar('SPLASH_ORCA_MODULE')):
        ut.red_exit(
            (
                'SPLASH_ORCA_MODULE environment variableis deprecated\n'
                'Please ee nimbus_splash documentation\n'
                'And set SPLASH_ORCA_VERSION instead'
            )
        )

    if uargs.instance in cfg.ORCA_SUPPORTED_INSTANCES:
        instance = uargs.instance
    else:
        ut.red_exit('Error: instance {uargs.instance} unsupported')

    # Read email environment variable
    try:
        email = os.environ['SPLASH_EMAIL']
    except KeyError:
        email = ''

    # Write job file
    for file in uargs.input_files:

        # Check input exists
        if not file.exists:
            ut.red_exit('Cannot locate {}'.format(file.name))

        # Check contents/format of input file and any file dependencies
        # with warnings.catch_warnings(record=True) as w:
        try:
            dependencies = job.parse_input_contents(
                file,
                instance_name=instance,
                skip_xyz=uargs.skip_xyz
            )
        except (ValueError, FileNotFoundError) as err:
            ut.red_exit(str(err))

            # if w:
            #     ut.cprint(str(w[0]), 'black_yellowbg')

        if uargs.verbose:
            print(dependencies)

        # Check for research allocation id environment variable
        raid = ut.get_envvar('SPLASH_RAID')
        if not len(raid):
            ut.red_exit('Please set SPLASH_RAID environment variable')

        job_file = job.write_file(
            file, instance, uargs.time, verbose=True,
            dependencies=list(dependencies.values()),
            orca_version=uargs.orca_version,
            research_allocation_id=raid,
            email=email
        )

        # Submit to queue
        if not uargs.no_start:
            subprocess.call('sbatch {}'.format(job_file), shell=True)

    return


def read_args(arg_list=None):
    '''
    Reader for command line arguments.
    Uses sub parsers for individual programs

    Parameters
    ----------
        args : argparser object
            command line arguments

    Returns
    -------
        None
    '''

    description = '''
    A package for working with Orca on Bath's Cloud HPC service
    '''

    epilog = '''
    To display options for a specific program, use splash \
    PROGRAMFILETYPE -h
    '''
    parser = argparse.ArgumentParser(
        description=description,
        epilog=epilog,
        formatter_class=argparse.RawDescriptionHelpFormatter
    )

    subparsers = parser.add_subparsers(dest='prog')

    submit = subparsers.add_parser(
        'submit',
        description='Generate Nimbus SLURM submission script'
    )
    submit.set_defaults(func=submit_func)

    submit.add_argument(
        'input_files',
        nargs='+',
        type=pathlib.Path,
        help='Orca input file name(s)'
    )

    default_instance = ut.get_envvar('SPLASH_DEFAULT_INSTANCE')
    if default_instance not in cfg.ORCA_SUPPORTED_INSTANCES:
        default_instance = 'spot-fsv2-16'

    submit.add_argument(
        '-i',
        '--instance',
        default=default_instance,
        type=str,
        help=f'Instance to run on, default is {default_instance}'
    )

    submit.add_argument(
        '-t',
        '--time',
        type=str,
        default='24:00:00',
        help='Time for job, formatted as HH:MM:SS, default 24:00:00'
    )

    default_version = ut.get_envvar('SPLASH_ORCA_VERSION')
    if not len(default_version):
        default_version = '6.0.1'

    submit.add_argument(
        '-o',
        '--orca_version',
        type=str,
        default=default_version,
        help='Version of orca to use e.g. 6.0.1'
    )

    submit.add_argument(
        '-sx',
        '--skip_xyz',
        action='store_true',
        help='Skip formatting check for .xyz file'
    )

    submit.add_argument(
        '-ns',
        '--no_start',
        action='store_true',
        help='If specified, jobs are not submitted to nimbus queue'
    )

    submit.add_argument(
        '-v',
        '--verbose',
        action='store_true',
        help='If specified, debug information is printed to screen'
    )

    submit.add_argument(
        '-ng',
        '--no_guess',
        action='store_true',
        help=(
            'If specified, gbw files found in results directory will not be'
            'used automatically'
        )
    )

    # If argument list is none, then call function func
    # which is assigned to help function
    parser.set_defaults(func=lambda user_args: parser.print_help())

    # read sub-parser
    _args, _ = parser.parse_known_args(arg_list)

    # select parsing option based on sub-parser
    if _args.prog in ['rst_opt']:
        args, job_args = parser.parse_known_args(arg_list)
        args.func(args, job_args)
    else:
        args = parser.parse_args(arg_list)
        args.func(args)
    return args


def interface():
    read_args()
    return



import pathlib
from orto import extractor as oe
import xyz_py as xyzp
import warnings

from . import utils as ut
from . import config as cfg


def write_file(input_file: str | pathlib.Path, instance_name: str, time: str,
               dependencies: dict[str, str | pathlib.Path],
               orca_version: str, research_allocation_id: str,
               verbose: bool = False, email: str = '',
               job_name: str = None) -> str:
    '''
    Writes slurm jobscript to file for ORCA calculation on nimbus

    Output file name is input_file with .slm extension

    Parameters
    ----------
    input_file : str | pathlib.Path
        Full path to input file, including extension
    instance_name : str
        Name of Nimbus instance to use e.g. spot-hc-44
    time : str
        Job time limit formatted as HH:MM:SS
    verbose : bool, default=False
        If True, prints job file name to screen
    dependencies : dict[str | pathlib.Path]
        Files required by job - e.g. xyz, gbw, hess
    orca_version: str
        string name of orca version
        e.g. 5.0.4
    research_allocation_id: str
        Research allocation ID
    email: str, optional
        If provided, adds the specified email to the jobscript.\n
        Users recieve an email for all changes in job status
    job_name: str, optional
        Name of job. If not provided, then job name defaults to stem of
        input file

    Returns
    -------
    str
        Name of jobscript file

    Raises
    ------
    ValueError
        If ORCA version not available on instance
    '''

    # Check instance's ORCA compatibility
    if instance_name not in cfg.ORCA_SUPPORTED_INSTANCES:
        raise ValueError(
            f'Specified instance {instance_name} does not support ORCA'
        )

    # Find relevant orca module for given instance
    try:
        orca_module = cfg.ORCA_MODULES[instance_name.split('-')[1]][orca_version] # noqa
    except KeyError:
        raise ValueError(
            f'ORCA version {orca_version} not available on instance {instance_name}' # noqa
        )

    # Get raw name of input file excluding path
    input_file = pathlib.Path(input_file)

    # Convert dependencies to Path objects
    dependencies: list[pathlib.Path] = [
        pathlib.Path(dep) for dep in dependencies
    ]

    # Name of job
    if job_name is None:
        job_name = input_file.stem

    job_file = pathlib.Path.joinpath(input_file.parent, f'{job_name}.slm')

    with open(job_file, 'w') as j:

        j.write('#!/bin/bash\n\n')

        j.write(f'#SBATCH --job-name={job_name}\n')
        j.write('#SBATCH --nodes=1\n')
        j.write('#SBATCH --ntasks-per-node={}\n'.format(
            cfg.INSTANCE_TOTAL_CORES[instance_name]
        ))
        j.write(f'#SBATCH --partition={instance_name}\n')
        j.write(f'#SBATCH --account={research_allocation_id}\n')
        j.write(f'#SBATCH --qos={instance_name}\n')
        j.write(f'#SBATCH --output={job_name}.%j.o\n')
        j.write(f'#SBATCH --error={job_name}.%j.e\n')

        if len(email):
            j.write(f'#SBATCH --mail-user={email}\n')
            j.write('#SBATCH --mail-type=ALL\n')

        j.write('#SBATCH --signal=B:USR1\n\n')

        j.write('# Job time\n')
        j.write(f'#SBATCH --time={time}\n\n')

        j.write('# name and path of the input/output files and locations\n')
        j.write(f'input={input_file}\n')
        j.write(f'output={job_name}.out\n')
        j.write(f'campaigndir={input_file.parent.absolute()}\n')
        j.write(f'results=$campaigndir/{input_file.stem}_results\n\n')

        j.write('# Local (Node) scratch, either node itself if supported ')
        j.write('or burstbuffer\n')
        j.write('if [ -d "/mnt/resource/" ]; then\n')
        j.write(
            '    localscratch="/mnt/resource/temp_scratch_$SLURM_JOB_ID"\n'
            '    mkdir $localscratch\n'
        )
        j.write('else\n')
        j.write('    localscratch=$BURSTBUFFER\n')
        j.write('fi\n\n')

        j.write('# If output file already exists, append OLD and ')
        j.write('last access time\n')
        j.write('if [ -f $output ]; then\n')
        j.write('    mv $output "$output"_OLD_$(date -r $output "+%Y-%m-%d-%H-%M-%S")\n') # noqa
        j.write('fi\n\n')

        j.write('# Copy files to localscratch\n')
        j.write('rsync -aP ')

        j.write(f'{input_file.absolute()}')

        for dep in dependencies:
            j.write(f' {dep.absolute()}')

        j.write(' $localscratch\n')
        j.write('cd $localscratch\n\n')

        j.write('# Write date and instance type to output\n')
        j.write('date > $campaigndir/$output\n')
        j.write('uname -n >> $campaigndir/$output\n\n')

        j.write('# Module system setup\n')
        j.write('source /apps/build/easy_build/scripts/id_instance.sh\n')
        j.write('source /apps/build/easy_build/scripts/setup_modules.sh\n\n')

        j.write('# Load orca\n')
        j.write('module purge\n')
        j.write(f'module load {orca_module}\n\n')

        j.write('# UCX transport protocols for MPI\n')
        j.write('export UCX_TLS=self,tcp,sm\n\n')

        j.write('# If timeout, evicted, cancelled, then manually end orca\n')

        j.write("trap 'echo signal recieved in BATCH!; kill -15 ")
        j.write('"${PID}"; wait "${PID}";')
        j.write("' SIGINT SIGTERM USR1 15\n\n")

        j.write('# Run calculation in background\n')
        j.write('# Catch the PID var for trap, and wait for process to end\n')
        j.write('$(which orca) $input >> $campaigndir/$output &\n')
        j.write('PID="$!"\n')
        j.write('wait "${PID}"\n\n')

        j.write('# Clean up and copy back files\n')

        j.write('# Check for existing results directory\n')
        j.write('cd $campaigndir\n')
        j.write('# If results directory already exists, append OLD and ')
        j.write('last access time\n')
        j.write('if [ -d $results ]; then\n')
        j.write(
            '    mv $results "$results"_OLD_$(date -r $results "+%Y-%m-%d-%H-%M-%S")\n') # noqa

        j.write('else\n')
        j.write('    mkdir $results\n')
        j.write('fi\n\n')
        j.write('cd $localscratch\n')

        j.write(
            'if compgen -G "$localscratch/*.res.Gradients" > /dev/null; then\n'
            '    rsync -aP --exclude=*.tmp* $localscratch/*.res.Gradients $results\n' # noqa
            'fi\n'
        )
        j.write(
            'if compgen -G "$localscratch/*.res.Dipoles" > /dev/null; then\n'
            '    rsync -aP --exclude=*.tmp* $localscratch/*.res.Dipoles $results\n' # noqa
            'fi\n'
        )
        j.write(
            'if compgen -G "$localscratch/*.res.Ramans" > /dev/null; then\n'
            '    rsync -aP --exclude=*.tmp* $localscratch/*.res.Ramans $results\n' # noqa
            'fi\n'
        )
        j.write(
            'if compgen -G "$localscratch/*.res.Nacmes" > /dev/null; then\n'
            '    rsync -aP --exclude=*.tmp* $localscratch/*.res.Nacmes $results\n' # noqa
            'fi\n'
        )

        j.write('rsync -aP --exclude=*.tmp* $localscratch/* $results\n')
        j.write('rm -r $localscratch\n')

    if verbose:
        if job_file.parent == pathlib.Path.cwd():
            ut.cprint(f'Submission script written to {job_file.name}', 'green')
        else:
            ut.cprint(f'Submission script written to {job_file}', 'green')

    return job_file


def parse_input_contents(input_file: str | pathlib.Path,
                         instance_name: str,
                         skip_xyz: bool = False) -> dict[str, pathlib.Path]:
    '''
    Checks contents of input file and returns file dependencies\n

    Parameters
    ----------
    input_file: str | pathlib.Path
        Full path to orca input file
    instance_name: str
        Name of Nimbus instance e.g. spot-fsv2-32
    skip_xyz: bool, default False
        If True, skips checking of xyz file formatting

    Returns
    -------
    dict[str, pathlib.Path]
        Relative paths of files required by this this input file\n
        Key is identifier (xyz, gbw, hess), Value is file name as Path object

    Raises
    ------
    ValueError
        If errors encountered in input file format or xyz file format
    FileNotFound
        If xyz, gbw, or hessian file cannot be found
    '''

    # Convert to Path object
    input_file = pathlib.Path(input_file)

    # Dependencies (files) of this input file
    dependencies = dict()

    # Load number of procs and amount of memory from orca input file
    try:
        n_procs = oe.NProcsInputExtractor.extract(input_file)[0]
    except oe.DataNotFoundError:
        raise ValueError(f'Missing Number of processors in {input_file}')

    try:
        maxcore = oe.MaxCoreInputExtractor.extract(input_file)[0]
    except oe.DataNotFoundError:
        raise ValueError(f'Missing Max Core Memory in {input_file}')

    # Check against selected instance
    if n_procs > cfg.INSTANCE_TOTAL_CORES[instance_name]:
        string = 'Error: Specified number of cores'
        string += f' {n_procs:d} in {input_file} exceeds '
        string += f'instance limit of {cfg.INSTANCE_TOTAL_CORES[instance_name]:d} cores' # noqa
        raise ValueError(string)

    # Check against selected instance
    if maxcore * n_procs > cfg.INSTANCE_TOTAL_MEM[instance_name]:
        string = 'Warning: Specified amount of memory'
        string += f' {maxcore:d} * {n_procs:d} in {input_file} exceeds '
        string += f'instance limit of {cfg.INSTANCE_TOTAL_MEM[instance_name]:d} MB' # noqa
        warnings.warn(string)

    # Get xyz file name and check it exists and is formatted correctly
    try:
        xyz_file = oe.XYZFileInputExtractor.extract(input_file)
    except oe.DataNotFoundError:
        xyz_file = []

    try:
        xyzline = oe.XYZInputExtractor.extract(input_file)
    except oe.DataNotFoundError:
        xyzline = []

    if not len(xyz_file) and not len(xyzline):
        raise ValueError(
            'Error: missing or incorrect *xyzfile or *xyz line in input'
        )

    if len(xyz_file) > 1 or len(xyzline) > 1 or len(xyz_file + xyzline) > 1: # noqa
        raise ValueError(
            'Error: multiple *xyzfile or *xyz lines in input. Only one can be present' # noqa
        )

    if len(xyz_file):
        xyz_file = pathlib.Path(xyz_file[0])
        if not xyz_file.is_file():
            raise FileNotFoundError(
                'Error: xyz file specified in input cannot be found'
            )
        dependencies['xyz'] = xyz_file

        if not skip_xyz:
            try:
                xyzp.check_xyz(
                    xyz_file.absolute(),
                    allow_indices=False
                )
            except xyzp.XYZError as e:
                raise ValueError(
                    f'{e}\n Use -sx to skip this check at your peril'
                )

    # Check if MORead and/or MOInp are present
    try:
        moread = oe.MOReadExtractor.extract(input_file)
    except oe.DataNotFoundError:
        moread = []
    try:
        moinp = oe.MOInpExtractor.extract(input_file)
    except oe.DataNotFoundError:
        moinp = []

    # Error if only one word present or if more than one of each word
    if len(moinp) ^ len(moread):
        raise ValueError('Error: Missing one of MOInp or MORead')
    if len(moinp) + len(moread) > 2:
        raise ValueError('Error: Multiple MORead and/or MOInp detected')

    if len(moinp):
        # Error if input orbitals have same stem as input file
        moinp = pathlib.Path(moinp[0])
        if moinp.stem == input_file.stem:
            raise ValueError(
                'Error: Stem of orbital and input files cannot match'
            )

        # Error if cannot find orbital file
        if not moinp.exists():
            raise FileNotFoundError(
                f'Error: Cannot find orbital file - {moinp}'
            )

        dependencies['gbw'] = moinp

    # Check if Hessname is present
    try:
        hess = oe.HessNameInputExtractor.extract(input_file)
    except oe.DataNotFoundError:
        hess = []

    if len(hess):
        hess = pathlib.Path(hess[0])
        if hess.stem == input_file.stem:
            raise ValueError(
                'Error: Stem of hessian and input files cannot match'
            )

        # Error if cannot find orbital file
        if not hess.exists():
            raise FileNotFoundError(
                f'Error: Cannot find Hessian file - {hess}'
            )

        dependencies['hess'] = hess

    return dependencies



import sys
import os


def red_exit(string):
    cprint(string, 'red')
    sys.exit(-1)
    return


def cstring(string: str, color: str) -> str:
    '''
    Returns colorised string

    Parameters
    ----------
    string: str
        String to print
    color: str {red, green, yellow, blue, magenta, cyan, white}
        String name of color

    Returns
    -------
    None
    '''

    ccodes = {
        'red': '\u001b[31m',
        'green': '\u001b[32m',
        'yellow': '\u001b[33m',
        'blue': '\u001b[34m',
        'magenta': '\u001b[35m',
        'cyan': '\u001b[36m',
        'white': '\u001b[37m',
        'black_yellowbg': '\u001b[30;43m\u001b[K',
        'white_bluebg': '\u001b[37;44m\u001b[K',
    }
    end = '\033[0m\u001b[K'

    # Count newlines at neither beginning nor end
    num_c_nl = string.rstrip('\n').lstrip('\n').count('\n')

    # Remove right new lines to count left new lines
    num_l_nl = string.rstrip('\n').count('\n') - num_c_nl
    l_nl = ''.join(['\n'] * num_l_nl)

    # Remove left new lines to count right new lines
    num_r_nl = string.lstrip('\n').count('\n') - num_c_nl
    r_nl = ''.join(['\n'] * num_r_nl)

    # Remove left and right newlines, will add in again later
    _string = string.rstrip('\n').lstrip('\n')

    _string = '{}{}{}{}{}'.format(l_nl, ccodes[color], _string, end, r_nl)

    return _string


def cprint(string: str, color: str):
    '''
    Prints colorised output to screen

    Parameters
    ----------
    string: str
        String to print
    color: str {red, green, yellow, blue, magenta, cyan, white}
        String name of color

    Returns
    -------
    None
    '''

    print(cstring(string, color))

    return


def get_envvar(var_str: str) -> str:
    '''
    Gets specified environment variable
    If undefined then returns empty string

    Parameters
    ----------
    var_str : str
        String name of environment variable

    Returns
    -------
    str
        Value of environment variable, or empty is not defined
    '''

    try:
        val = os.environ[var_str]
    except KeyError:
        val = ''

    return val


def flatten_recursive(to_flat: list[list]) -> list:
    '''
    Flatten a list of lists recursively.

    Parameters
    ----------
    to_flat: list

    Returns
    -------
    list
        Input list flattened to a single list
    '''

    if to_flat == []:
        return to_flat
    if isinstance(to_flat[0], list):
        return flatten_recursive(to_flat[0]) + flatten_recursive(to_flat[1:])
    return to_flat[:1] + flatten_recursive(to_flat[1:])



import setuptools

long_description = '''
Splash - the sound an Orca would make if it had a Bath.\n\n

`nimbus_splash`, or `splash` for short, is a package to make life easier
when using the University of Bath's `Nimbus` cloud computing suite for Orca
calculations.\n\n

Please see the `nimbus_splash` documentation for more details.
'''

# DO NOT EDIT THIS NUMBER!
# IT IS AUTOMATICALLY CHANGED BY python-semantic-release
__version__ = '1.8.2'

setuptools.setup(
    name='Nimbus Splash',
    version=__version__,
    author='Jon Kragskow',
    author_email='jgck20@bath.ac.uk',
    description=(
        'A package to make life easier when using the University of '
        'Bath\'s cloud computing suite for Orca calculations.'
    ),
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/jonkragskow/nimbus_splash',
    project_urls={
        'Bug Tracker': 'https://github.com/JonKragskow/nimbus_splash/issues',
        'Documentation': 'https://www.kragskow.dev/nimbus_splash/index.html'
    },
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
    ],
    package_dir={'': '.'},
    packages=setuptools.find_packages(),
    python_requires='>=3.10',
    install_requires=['numpy>=2.1.3', 'xyz_py>=5.13.1', 'orto>=0.10.0'],
    entry_points={
        'console_scripts': [
            'splash = nimbus_splash.cli:interface',
            'nimbussplash = nimbus_splash.cli:interface',
            'nimbus_splash = nimbus_splash.cli:interface'
        ]
    }
)
